-- MODULE:
--   GoldParser
--
-- PURPOSE:
--   A Haskell engine for the GOLD Parsing System
--   <http://www.devincook.com/goldparser/>.
--
-- AUTHOR:
--   Stephen Thompson <stephen@solarflare.org.uk>
--
-- COPYRIGHT:
--   Copyright (C) Stephen Thompson, 2011.
--
--   This file is part of Knights Map Editor.
--
--   Knights Map Editor is free software: you can redistribute it
--   and/or modify it under the terms of the GNU General Public
--   License as published by the Free Software Foundation, either
--   version 3 of the License, or (at your option) any later version.
--
--   Knights Map Editor is distributed in the hope that it will be
--   useful, but WITHOUT ANY WARRANTY; without even the implied
--   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
--   See the GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with Knights Map Editor. If not, see
--   <http://www.gnu.org/licenses/>.

module GoldParser 
    ( loadCgt, 
      initialParseState, 
      parse,
      ParseState,
      injectStream,
      ParseMessage(Accept,Reduce,Shift,SyntaxError,LexicalError,CommentError),
      Symbol(SError,SCommentStart,SCommentEnd,SEof,SWhitespace,SCommentLine,Sym)
    )
where


-- TODO before releasing this ...
---- Change the error messages? Or consider whether "error" should be used at all (it's probably ok)?
---- Comment handling


import Control.Monad.State
import Data.List
import qualified Data.Map as Map
import Data.Maybe

--
-- Symbol
--

data Symbol = SError | SCommentStart | SCommentEnd | SEof | SWhitespace | SCommentLine | Sym Int deriving Eq


--
-- DfaState, DfaEdge
--

data DfaState = DfaState { acceptSymbol :: Maybe Symbol,
                           edges        :: [DfaEdge]     }

data DfaEdge = DfaEdge { acceptedChars :: [Char],
                         targetState   :: DfaState }

emptyDfaState = DfaState Nothing []


-- findEdgeContainingChar:
-- look for the edge that accepts a given char (from a given state)

findEdgeContainingChar :: DfaState -> Char -> Maybe DfaEdge
findEdgeContainingChar state searchChar
    = findEdgeHelp searchChar (edges state)

findEdgeHelp :: Char -> [DfaEdge] -> Maybe DfaEdge
findEdgeHelp searchChar [] = Nothing
findEdgeHelp searchChar (e:es) 
    = if searchChar `elem` (acceptedChars e) then Just e
      else findEdgeHelp searchChar es 

-- readToken: 
-- Read one token from the source stream
-- Returns symbol, string representation and remaining stream. (or Nothing on error.)

-- Note: We don't bother too much with error handling here. It may be possible however to return a counter of
-- newlines that have been read; at least then we would know the line number of an error. 

readToken :: String -> DfaState -> Maybe (Symbol, String, String)
readToken [] _ = Just (SEof, "", [])        -- EOF here is a token in its own right
readToken stream state = readTokenHelp stream state  -- Otherwise call the "real" readToken function

readTokenHelp :: String -> DfaState -> Maybe (Symbol, String, String)
readTokenHelp input state = 
    -- Find whether any edge accepts the next character on the input stream.
    let edge = case input of
                 []     -> Nothing                        -- (EOF, so no edge can accept the next char, by definition.)
                 (x:xs) -> findEdgeContainingChar state x
    in case edge of
         Just e -> 
             -- Edge e accepts the character on the stream.
             -- Transition to the target state of e
             case readTokenHelp (tail input) (targetState e) of
               Just (sym, tok, stream) -> Just (sym, ((head input):tok), stream)  -- Success (attach x to front of tok).
               Nothing -> Nothing                                                 -- Failure (propagate the error upwards).
         Nothing -> 
             -- No edge accepts the character on the stream
             -- If current state allows an "accept symbol" then we can return that,
             -- else it's an error.
             case (acceptSymbol state) of Just sym -> Just (sym, [], input)
                                          Nothing -> Nothing




--
-- Parser
--

data Rule = Rule { ruleIndex :: Int,
                   headSymbol :: Symbol,
                   symbolCount :: Int    -- no of stack entries to pop while reducing this rule
                 }

data ActionType = AShift LalrState 
                | AReduce Rule 
                | AAccept

data LalrAction = LalrAction { actionSymbol :: Symbol,
                               actionType   :: ActionType }

newtype LalrState = LalrState { stateActions :: [LalrAction] }

emptyLalrState = LalrState []

data ParseResult = PShift 
                 | PReduce Int   -- the rule index for the reduction is included
                 | PAccept 
                 | PSyntaxError 

getActionFor :: Symbol -> LalrState -> Maybe LalrAction
getActionFor symbol state = find (\x -> (actionSymbol x) == symbol) (stateActions state)

-- The parseToken function

parseToken :: Symbol -> String -> [LalrState] -> (ParseResult, [LalrState])
parseToken _ _ [] = error "gold parser error -- parse stack was empty in parseToken"
parseToken symbol token stack@(state:states) =
    case (getActionFor symbol state) of
      Nothing -> (PSyntaxError, stack)
      Just action -> 
          case (actionType action) of 
            AAccept            -> (PAccept, stack)
            AShift targetState -> (PShift, (targetState:stack))
            AReduce reduceRule -> let reducedStack = drop (symbolCount reduceRule) stack
                                      reduction    = ruleIndex reduceRule 
                                      newAction    = fromJust $ getActionFor (headSymbol reduceRule) (head reducedStack)
                                  in case (actionType newAction) of 
                                       AShift newState ->
                                           let newStack = newState : reducedStack
                                           in (PReduce reduction, newStack)
                                       _ -> error "invalid action type for reduction"

--
-- CGT Loading
--

-- An intermediate structure to hold the various tables in the CGT file.
data CGTData = CGTData { characterSetTable :: Map.Map Int String,
                         symbolTable       :: Map.Map Int Symbol,
                         ruleTable         :: Map.Map Int Rule,
                         dfaStateTable     :: Map.Map Int DfaState,
                         lalrStateTable    :: Map.Map Int LalrState,
                         caseSensitive     :: Bool,
                         initialDfaState   :: DfaState,
                         initialLalrState  :: LalrState
                       }

blankCGTData = CGTData Map.empty Map.empty Map.empty Map.empty Map.empty False emptyDfaState emptyLalrState

-- Set this up as a state monad ...
-- The [Char] component of the state represents the remaining input stream. This changes as we read data.
-- The CGTData component of the state represents the final loaded CGT. Since we don't actually know this yet :) we are relying 
-- on lazy evaluation for this to work. (It's used for looking up the dfa/lalr states from their integer index codes; in this 
-- way we can build circular data structures from an input stream.)

type CGTLoad a = State (CGTData,[Char]) a

getFinalCGT :: CGTLoad CGTData
getFinalCGT = do
  (x,_) <- get
  return x

getStream :: CGTLoad [Char]
getStream = do
  (_,x) <- get
  return x

putStream :: [Char] -> CGTLoad ()
putStream newStr = do
  (x,y) <- get
  put (x, newStr)

atEof :: CGTLoad Bool
atEof = do
  x <- getStream
  return (null x)

-- Actually this doesn't really read unicode... we just ignore every 2nd byte :)
readUnicodeString :: CGTLoad String
readUnicodeString = do
  str <- getStream
  case str of
    ('\0':(_:xs)) -> do putStream xs
                        return ""
    (x:(_:xs)) -> do putStream xs
                     remainder <- readUnicodeString
                     return (x:remainder)
    _ -> error "bad cgt: unicode string not properly terminated"
         
readChar :: CGTLoad Char
readChar = do
  str <- getStream
  case str of 
    (x:xs) -> do putStream xs
                 return x
    _ -> error "bad cgt: unexpected end of file"

readUshort :: CGTLoad Int
readUshort = do
  str <- getStream
  case str of
    (x:(y:xs)) -> do putStream xs
                     return (hi*256 + lo)
                         where lo = fromEnum x
                               hi = fromEnum y
    _ -> error "bad cgt: unexpected end of file"


-- Entry readers

readByteEntry :: CGTLoad Int
readByteEntry = do
  str <- getStream
  case str of 
    ('b':(x:xs)) -> do putStream xs
                       return (fromEnum x)
    _ -> error "bad cgt: error reading byte entry"

readBooleanEntry :: CGTLoad Bool
readBooleanEntry = do
  str <- getStream
  case str of 
    ('B':('\0':xs)) -> do putStream xs
                          return False
    ('B':('\1':xs))  -> do putStream xs
                           return True
    _ -> error "bad cgt: error reading boolean entry"

readIntegerEntry :: CGTLoad Int
readIntegerEntry = do
  str <- getStream
  case str of 
    ('I':(x:(y:xs))) -> do putStream xs
                           let lo = fromEnum x
                               hi = fromEnum y
                               res = 256*hi + lo
                           if res > 32767 then return (65536 - res)  -- result should be signed
                              else return res
    _ -> error "bad cgt: error reading integer entry"

readStringEntry :: CGTLoad String
readStringEntry = do
  str <- getStream
  case str of 
    ('S':xs) -> do putStream xs
                   readUnicodeString
    _ -> error "bad cgt: error reading string entry"

readEmptyEntry :: CGTLoad ()
readEmptyEntry = do
  str <- getStream
  case str of 
    ('E':xs) -> putStream xs
    _ -> error "bad cgt: error reading empty entry"


-- Functions to load the actual CGT data

readCGTData :: CGTLoad CGTData
readCGTData = do
  header <- readUnicodeString
  when (header /= "GOLD Parser Tables/v1.0") (error "bad cgt: invalid file header")
  readRecords

data RecordResult = CharacterSetTableEntry Int String
                  | SymbolTableEntry Int Symbol
                  | RuleTableEntry Int Rule
                  | DfaStateTableEntry Int DfaState
                  | LalrStateTableEntry Int LalrState
                  | CaseSensitive Bool
                  | InitialStateRecord DfaState LalrState
                  | NullRecord

-- This just loops over all records. 
readRecords :: CGTLoad CGTData
readRecords = do
  eof <- atEof
  if eof then return blankCGTData
    else do thisRecord <- readRecord
            remainder <- readRecords
            case thisRecord of
              -- Here we update "remainder" to include the new record. 
              CharacterSetTableEntry k v -> return remainder { characterSetTable = Map.insert k v (characterSetTable remainder) }
              SymbolTableEntry k v -> return remainder { symbolTable = Map.insert k v (symbolTable remainder) }
              RuleTableEntry k v -> return remainder { ruleTable = Map.insert k v (ruleTable remainder) }
              DfaStateTableEntry k v -> return remainder { dfaStateTable = Map.insert k v (dfaStateTable remainder) }
              LalrStateTableEntry k v -> return remainder { lalrStateTable = Map.insert k v (lalrStateTable remainder) }
              CaseSensitive foo -> return remainder { caseSensitive = foo }
              InitialStateRecord initDfa initLalr -> return remainder { initialDfaState = initDfa, initialLalrState = initLalr }
              NullRecord -> return remainder


-- Read an individual record. Return a "RecordResult"
readRecord :: CGTLoad RecordResult
readRecord = do
  -- Type of record -- should be 'M'
  rtype <- readChar
  when (rtype /= 'M') (error "bad cgt: illegal record type")
         
  -- Number of entries in this record
  nEntries <- readUshort
  when (nEntries < 1) (error "bad cgt: illegal number of entries")
  
  -- First entry tells us what kind of record this is
  rtype' <- readByteEntry
  let rtype'' :: Char
      rtype'' = toEnum rtype'

  case rtype'' of
    'P' -> readParameters 
    'T' -> readTableCounts
    'C' -> readCharacterSetTableEntry 
    'S' -> readSymbolTableEntry 
    'R' -> readRuleTableEntry nEntries 
    'I' -> readInitialStateRecord 
    'D' -> readDfaState nEntries
    'L' -> readLalrState nEntries
    _   -> error "bad cgt: illegal hunk type"


readParameters :: CGTLoad RecordResult
readParameters = do
  readStringEntry  -- name
  readStringEntry  -- version
  readStringEntry  -- author
  readStringEntry  -- about
  caseSen <- readBooleanEntry
  readIntegerEntry  -- start symbol
  return (CaseSensitive caseSen)

readTableCounts :: CGTLoad RecordResult
readTableCounts = do
  readIntegerEntry  -- symbol table size
  readIntegerEntry  -- character set table size
  readIntegerEntry  -- rule table size
  readIntegerEntry  -- dfa state table size
  readIntegerEntry  -- lalr state table size
  return NullRecord  -- We actually ignore the table counts record here :)

readCharacterSetTableEntry :: CGTLoad RecordResult
readCharacterSetTableEntry = do
  i <- readIntegerEntry
  x <- readStringEntry
  return $ CharacterSetTableEntry i x

readSymbolTableEntry :: CGTLoad RecordResult
readSymbolTableEntry = do
  i <- readIntegerEntry   -- Index
  readStringEntry         -- Symbol name (ignored)
  x <- readIntegerEntry   -- Symbol Kind, which we translate here into one of our Symbol codes.
  if x >= 2 && x <= 7 then 
      let sym = case x of 
                  2 -> SWhitespace
                  3 -> SEof
                  4 -> SCommentStart
                  5 -> SCommentEnd
                  6 -> SCommentLine
                  7 -> SError
      in return $ SymbolTableEntry i sym
    else return NullRecord  -- Terminals and nonterminals aren't added to the symbol table.

readRuleTableEntry :: Int -> CGTLoad RecordResult
readRuleTableEntry nEntries = do
  index   <- readIntegerEntry  -- Index
  headSym <- readIntegerEntry  -- Non-terminal symbol
  readEmptyEntry               -- Reserved
  when (nEntries < 4) (error "bad cgt: illegal number of entries for rule")
  mapM_ (\_->readIntegerEntry) [1..(nEntries-4)]   -- Discard Symbols.
  return $ RuleTableEntry index (Rule { ruleIndex   = index,
                                        headSymbol  = Sym headSym,
                                        symbolCount = (nEntries-4)
                                      } )

readInitialStateRecord :: CGTLoad RecordResult
readInitialStateRecord = do
  initDfa  <- readIntegerEntry
  initLalr <- readIntegerEntry
  finalCGT <- getFinalCGT
  let initDfaState = Map.lookup initDfa (dfaStateTable finalCGT)
      initLalrState = Map.lookup initLalr (lalrStateTable finalCGT)
  return $ InitialStateRecord (fromJust initDfaState) (fromJust initLalrState)

readDfaState :: Int -> CGTLoad RecordResult
readDfaState nEntries = do
  finalCGT <- getFinalCGT

  index <- readIntegerEntry
  acceptState <- readBooleanEntry
  acceptSymbolIndex <- readIntegerEntry
  readEmptyEntry -- reserved

  let acceptSymbol = if acceptState then 
                         let maybeSymbol = Map.lookup acceptSymbolIndex (symbolTable finalCGT)
                         in case maybeSymbol of 
                              Just foo -> Just foo
                              Nothing -> Just (Sym acceptSymbolIndex)
                     else Nothing

  edges <- mapM (readEdge finalCGT) [1..(nEntries-5) `div` 3]

  return $ DfaStateTableEntry index (DfaState acceptSymbol edges)

      where readEdge :: CGTData -> a -> CGTLoad DfaEdge
            readEdge finalCGT _ = do
              charsetIndex <- readIntegerEntry
              targetStateIndex <- readIntegerEntry
              readEmptyEntry -- reserved
              let charset = Map.lookup charsetIndex (characterSetTable finalCGT) 
                  targetState = Map.lookup targetStateIndex (dfaStateTable finalCGT)
              return $ DfaEdge (fromJust charset) (fromJust targetState)


readLalrState :: Int -> CGTLoad RecordResult
readLalrState nEntries = do
  finalCGT <- getFinalCGT

  index <- readIntegerEntry
  readEmptyEntry -- reserved

  actions <- mapM (readAction finalCGT) [1..(nEntries-3) `div` 4]
  return $ LalrStateTableEntry index (LalrState actions)

      where readAction :: CGTData -> a -> CGTLoad LalrAction
            readAction finalCGT _ = do
              symbolCode <- readIntegerEntry
              actionTypeCode <- readIntegerEntry
              actionTargetCode <- readIntegerEntry
              readEmptyEntry

              let shiftTarget = case Map.lookup actionTargetCode (lalrStateTable finalCGT) of
                                  Just lalrState -> lalrState
                                  Nothing -> error "bad cgt: bad lalr state code"
                  reduceRule  = case Map.lookup actionTargetCode (ruleTable finalCGT) of 
                                  Just rule -> rule
                                  Nothing -> error "bad cgt: bad reduction rule code"
            
              let actionType = case actionTypeCode of 
                                 1 -> AShift shiftTarget
                                 2 -> AReduce reduceRule
                                 3 -> AShift shiftTarget
                                 4 -> AAccept

              let testSymbol = case (Map.lookup symbolCode (symbolTable finalCGT)) of
                                 Just foo -> foo
                                 Nothing -> (Sym symbolCode)

              return $ LalrAction testSymbol actionType



-- Function to actually load CGT data from a stream (list of Chars).
-- Remember to open the file in binary mode !!

type CGT = (DfaState, LalrState)

loadCgt :: [Char] -> CGT
loadCgt inputByteStream = let (result, _) = runState readCGTData (result, inputByteStream)
                          in (initialDfaState result, initialLalrState result)


--
-- Parse Function
--

data ParseMessage = Reduce Int           -- carries the rule number
                  | Shift Symbol String  -- carries the symbol code and the token itself
                  | Accept 
                  | LexicalError 
                  | SyntaxError
                  | CommentError

-- ParseState holds the string being parsed, the Lalr state stack, and
-- the initial DFA state (this is the same for each token read), plus a
-- "current token" (possibly).

data ParseState = ParseState { dfaState   :: DfaState,
                               stream     :: String,
                               parseStack :: [LalrState],
                               curTok     :: Maybe (Symbol,String)
                             }

initialParseState :: CGT -> String -> ParseState
initialParseState cgt str = ParseState { dfaState   = fst cgt,
                                         stream     = str,
                                         parseStack = [snd cgt],
                                         curTok     = Nothing
                                       }

-- This file can be used to inject "includes" into the current stream.
injectStream :: ParseState -> [Char] -> ParseState
injectStream state includedFile = state { stream = includedFile ++ (stream state) }

-- Top level parse function
-- Input: parse state
-- Output: new parse state plus ParseMessage

parse :: ParseState -> (ParseMessage, ParseState)
parse state = 
    case (curTok state) of 
      Nothing -> let newState = getNextToken state
                 in case (curTok newState) of 
                      Nothing -> (LexicalError, newState)
                      Just tok -> parse' newState tok
      Just tok -> parse' state tok

getNextToken :: ParseState -> ParseState
getNextToken state =
    let tokResult = readToken (stream state) (dfaState state)
    in case tokResult of 
         Nothing -> state { stream = tail (stream state) } -- Lexical error (drop one char in this case)
         Just (sym, tok, newStream) -> state { stream = newStream, curTok = Just (sym, tok) }   -- OK

-- This function does the actual parsing (once we know we have a token available).
-- TODO: Comment mode has not been implemented. 
parse' :: ParseState -> (Symbol, String) -> (ParseMessage, ParseState)
parse' state (sym,tok) = 
    case sym of 
      SWhitespace -> parse (state {curTok = Nothing})                        -- Filter out whitespace tokens
      SCommentLine -> parse (state {curTok = Nothing, stream = discardLine}) -- Filter out line comments
             where discardLine = dropWhile (/='\10') (stream state)      -- 10 = new line character

      _ -> let (parseResult, newStack) = parseToken sym tok (parseStack state)
               parseMessage = case parseResult of
                                PAccept -> Accept 
                                PReduce r -> Reduce r
                                PShift -> Shift sym tok
                                PSyntaxError -> SyntaxError  -- TODO do we need to remove cur tok?
               newState' = case parseResult of 
                             PShift -> state { curTok = Nothing }   -- Must drop current token after a shift
                             _ -> state                             -- otherwise keep it
               newState  = newState' { parseStack = newStack }
           in (parseMessage, newState)


